@@ -1,6 +1,6 @@
 package Net::SFTP::Foreign::Helpers;
 
-our $VERSION = '1.41';
+our $VERSION = '1.74_06';
 
 use strict;
 use warnings;
@@ -12,29 +12,64 @@ use Scalar::Util qw(tainted);
 
 require Exporter;
 our @ISA = qw(Exporter);
-our @EXPORT = qw( _do_nothing
-		  _sort_entries
+our @EXPORT = qw( _sort_entries
 		  _gen_wanted
 		  _ensure_list
-		  _glob_to_regex
-                  _tcroak
                   _catch_tainted_args
                   _debug
                   _gen_converter
+		  _hexdump
+		  $debug
                 );
+our @EXPORT_OK = qw( _is_lnk
+                     _is_dir
+                     _is_reg
+                     _do_nothing
+		     _glob_to_regex
+                     _file_part
+                     _umask_save_and_set
+                     _tcroak );
 
-sub _do_nothing {}
+our $debug;
+
+BEGIN {
+    eval "use Time::HiRes 'time'"
+	if ($debug and $debug & 256)
+}
 
 sub _debug {
-    local $\;
-    if ($Net::SFTP::Foreign::debug & 256) {
-        print STDERR "#", $$, " ", @_,"\n"
+    local ($\, $!);
+    my $caller = '';
+    if ( $debug & 8192) {
+	$caller = (caller 1)[3];
+	$caller =~ s/[\w:]*:://;
+	$caller .= ': ';
+    }
+    if ($debug & 256) {
+	my $ts = sprintf("%010.5f", time);
+        print STDERR "#$$ $ts $caller", @_,"\n"
     }
     else {
-        print STDERR '# ', @_,"\n"
+        print STDERR "# $caller", @_,"\n"
     }
 }
 
+sub _hexdump {
+    local ($\, $!);
+    no warnings qw(uninitialized);
+    my $data = shift;
+    while ($data =~ /(.{1,32})/smg) {
+        my $line=$1;
+        my @c= (( map { sprintf "%02x",$_ } unpack('C*', $line)),
+                (("  ") x 32))[0..31];
+        $line=~s/(.)/ my $c=$1; unpack("c",$c)>=32 ? $c : '.' /egms;
+	local $\;
+        print STDERR join(" ", @c, '|', $line), "\n";
+    }
+}
+
+sub _do_nothing {}
+
 {
     my $has_sk;
     sub _has_sk {
@@ -84,8 +119,8 @@ sub _ensure_list {
     local $@;
     local $SIG{__DIE__};
     local $SIG{__WARN__};
-    return @$l if eval { @$l >= 0 };
-    return ($l);
+    no warnings;
+    (eval { @$l; 1 } ? @$l : $l);
 }
 
 sub _glob_to_regex {
@@ -142,7 +177,6 @@ sub _glob_to_regex {
 		    else {
 			croak "invalid glob pattern";
 		    }
-				
 		}
 		else {
                     $wildcards--;
@@ -175,7 +209,7 @@ sub _catch_tainted_args {
     my $i;
     for (@_) {
         next unless $i++;
-        if (tainted $_) {
+        if (tainted($_)) {
             my (undef, undef, undef, $subn) = caller 1;
             my $msg = ( $subn =~ /::([a-z]\w*)$/
                         ? "Insecure argument '$_' on '$1' method call"
@@ -183,31 +217,29 @@ sub _catch_tainted_args {
             _tcroak($msg);
         }
         elsif (ref($_)) {
-            local $@;
-            local $SIG{__DIE__};
-            for (eval { values %$_ }) {
-                if (tainted $_) {
-                    my (undef, undef, undef, $subn) = caller 1;
-                    my $msg = ( $subn =~ /::([a-z]\w*)$/
-                                ? "Insecure argument on '$1' method call"
-                                : "Insecure argument on method call" );
-                    _tcroak($msg);
-                }
+            for (grep tainted($_),
+		 do { local ($@, $SIG{__DIE__}); eval { values %$_ }}) {
+		my (undef, undef, undef, $subn) = caller 1;
+		my $msg = ( $subn =~ /::([a-z]\w*)$/
+			    ? "Insecure argument on '$1' method call"
+			    : "Insecure argument on method call" );
+		_tcroak($msg);
             }
         }
     }
 }
 
 sub _gen_dos2unix {
+    my $unix2dos = shift;
+    my $name = ($unix2dos ? 'unix2dos' : 'dos2unix');
     my $previous;
     my $done;
     sub {
-        $done and die "Internal error: bad calling sequence for unix2dos transformation";
-        my $debug = ($Net::SFTP::Foreing::debug and $Net::SFTP::Foreing::debug & 128);
+        $done and die "Internal error: bad calling sequence for $name transformation";
         my $adjustment = 0;
         for (@_) {
-            if ($debug) {
-                _debug ("before dos2unixunix2dos: previous: $previous, data follows...");
+            if ($debug and $debug & 128) {
+                _debug ("before $name: previous: $previous, data follows...");
                 _hexdump($_);
             }
             if (length) {
@@ -216,7 +248,12 @@ sub _gen_dos2unix {
                     $_ = "\x0d$_";
                 }
                 $adjustment -= $previous = s/\x0d\z//s;
-                $adjustment -= s/\x0d\x0a/\x0a/gs;
+                if ($unix2dos) {
+                    $adjustment += s/(?<!\x0d)\x0a/\x0d\x0a/gs;
+                }
+                else {
+                    $adjustment -= s/\x0d\x0a/\x0a/gs;
+                }
             }
             elsif ($previous) {
                 $previous = 0;
@@ -224,8 +261,8 @@ sub _gen_dos2unix {
                 $adjustment++;
                 $_ = "\x0d";
             }
-            if ($debug) {
-                _debug ("after dos2unix: previous: $previous, adjustment: $adjustment, data follows...");
+            if ($debug and $debug & 128) {
+                _debug ("after $name: previous: $previous, adjustment: $adjustment, data follows...");
                 _hexdump($_);
             }
             return $adjustment;
@@ -233,22 +270,6 @@ sub _gen_dos2unix {
     }
 }
 
-sub _unix2dos {
-    my $debug = ($Net::SFTP::Foreing::debug and $Net::SFTP::Foreing::debug & 128);
-    if ($debug) {
-        _debug ("before unix2dos: data follows...");
-        _hexdump($_[0]);
-    }
-    my $adjustment = $_[0] =~ s/\x0a/\x0d\x0a/gs;
-    if ($debug) {
-        _debug ("before unix2dos: adjustment: $adjustment, data follows...");
-        _hexdump($_[0]);
-    }
-    $adjustment;
-}
-
-sub _gen_unix2dos { \&_unix2dos }
-
 sub _gen_converter {
     my $conversion = shift;
 
@@ -259,7 +280,7 @@ sub _gen_converter {
             return sub {
                 my $before = length $_[0];
                 $conversion->($_[0]);
-                length $_[0] - $before;
+                length($_[0]) - $before;
             }
         }
         else {
@@ -267,15 +288,36 @@ sub _gen_converter {
         }
     }
     elsif ($conversion eq 'dos2unix') {
-        return _gen_dos2unix;
+        return _gen_dos2unix(0);
     }
     elsif ($conversion eq 'unix2dos') {
-        return _gen_unix2dos
+        return _gen_dos2unix(1);
     }
     else {
         croak "unknown conversion '$conversion'";
     }
 }
 
+sub _is_lnk { (0120000 & shift) == 0120000 }
+sub _is_dir { (0040000 & shift) == 0040000 }
+sub _is_reg { (0100000 & shift) == 0100000 }
+
+sub _file_part {
+    my $path = shift;
+    $path =~ m{([^/]*)$} or croak "unable to get file part from path '$path'";
+    $1;
+}
+
+sub _umask_save_and_set {
+    my $umask = shift;
+    if (defined $umask) {
+        my $old = umask $umask;
+        return bless \$old, 'Net::SFTP::Foreign::Helpers::umask_saver';
+    }
+    ()
+}
+
+sub Net::SFTP::Foreign::Helpers::umask_saver::DESTROY { umask ${$_[0]} }
+
 1;
 
